/* Mixer930 0.2/REXX (Jul 18 1999) - OPTi930 commandline/textmode mixer
 * by Anton Dushko (2:5030/611.9)
 * This stuff is FREE. Comments & suggestions are welcome.
 */

/* Banner */
Say "Mixer930 0.2/REXX (Jul 18 1999) - OPTi930 commandline/textmode mixer" || "0D0A"x ||,
    "by Anton Dushko (2:5030/611.9)" || "0D0A"x

/* Startup operations */
Trace Off
Address CMD
Call On Halt Name CtrlBreak
Call RxFuncAdd "SysLoadFuncs", "RexxUtil", "SysLoadFuncs"
Call SysLoadFuncs

/* Load port IO rotunes */
Call RxFuncAdd "RxLoadInOut", "RexxIo", "RxLoadInOut"
If RxFuncQuery("RxLoadInOut") = 1 Then
Do
  Say "Port IO rotunes (REXXIO.DLL) not avialable"
  Exit -2
End
Call RxLoadInOut

/* Initialize mixer */
If MixerInit() = 0 Then
Do
  Say "No OPTi930-based sound card found"
  Exit -1
End

/* Get arguments */
Parse Upper Arg Params

/* Defaults */
ChangeWindowSize = 0
InteractiveMode = 0
CommandLineMode = 0

/* Process switches */
Do I = 1 To Words(Params)
  Switch = Word(Params, I)
  If Left(Switch, 1) = "/" | Left(Switch, 1) = "-" Then
  Do
    /* Process switch */
    Switch = SubStr(Switch, 2)
    Select
      /* Help */
      When Switch = "?" | Switch = "H" Then
      Do
        Say "Usage: Mixer930 [switches] [mixer controls]" || "0D0A"x ||,
            "  Switches:" || "0D0A"x ||,
            "    /{H|?} - This help" || "0D0A"x ||,
            "    /W     - Change window size to match mixer's interface" || "0D0A"x ||,
            "    /I     - Force interactive mode" || "0D0A"x ||,
            "  Controls syntax:" || "0D0A"x ||,
            "    Control:              <device>={<setup>|<left setup>/<right setup>}" || "0D0A"x ||,
            "    Device:               {Main|Wave|MIDI|CD|Line|Mic|Rec}" || "0D0A"x ||,
            "    Setup:                <setting>[,<setting>[,<setting>]]" || "0D0A"x ||,
            "    Setting (except rec): {<volume>|{M|U}}" || "0D0A"x ||,
            "    Setting (recording):  {<volume>|{L|C|M|X}|{G|N}}" || "0D0A"x ||,
            "    Volume:               {<absolute>|<db>dB|<percentage>%}" || "0D0A"x ||,
            "  If no controls specified, mixer starts in interactive mode:" || "0D0A"x ||,
            "    Cursor keys       - Control selection" || "0D0A"x ||,
            "    Numpad 7/1        - Increase/descrease left channel volume" || "0D0A"x ||,
            "    Ctrl + Numpad 7/1 - Set max/min volume on left channel" || "0D0A"x ||,
            "    Numpad 4          - Mute/unmute left channel (source selection for" || "0D0A"x ||,
            "                        recording)" || "0D0A"x ||,
            "    Ctrl + Numpad 4   - Left channel +20dB microphone gain toggle (recording" || "0D0A"x ||,
            "                        only)" || "0D0A"x ||,
            "    Numpad 9/3/6      - Same for right channel" || "0D0A"x ||,
            "    Numpad 8/5/2      - Same for both channels" || "0D0A"x ||,
            "    Numpad 0/.        - Propagate left/right channel settings to both channels" || "0D0A"x ||,
            "    Space             - Refresh display"
        Exit 1
      End
      /* Resize window */
      When Switch = "W" Then
        ChangeWindowSize = 1
      /* Force interactive mode */
      When Switch = "I" Then
        InteractiveMode = 1
      /* Ignore others */
      Otherwise
        NOP
    End
  End
  Else
  Do
    /* At least one control found - switch to commandline mode */
    Params = SubWord(Params, I)
    CommandLineMode = 1
    Leave
  End
End I

/* Select mode */
If CommandLineMode Then
Do
  Call CLProc Params
  If InteractiveMode Then
    Call UIProc ChangeWindowSize
End
Else
  Call UIProc ChangeWindowSize

/* All OK */
Exit 0


/* Command line */

/* Commandline mode procedure */
CLProc:
  Arg Params

  /* Initialize CL data */
  Call CLInit

  /* Header */
  If \InteractiveMode Then
    Call CLDrawMixerFace

  /* Main loop */
  Do I = 1 To Words(Params)
    Parse Value Word(Params, I) With DeviceName"="LeftSetup"/"RightSetup

    /* Skip this control if setup is empty */
    If LeftSetup = "" Then 
      Iterate

    /* Find deivce index */ 
    CC = 0
    Do J = 1 To MixerControls.Num
      If MixerControls.J.Abbrev = DeviceName Then
      Do
        CC = J
        Leave
      End
    End J

    /* Skip this control if device index not found */
    If CC = 0 Then
      Iterate
    
    /* Convert commas to spaces */
    LeftSetup = Translate(LeftSetup, " ", ",")
    RightSetup = Translate(RightSetup, " ", ",")

    /* Read mixer settings */
    Interpret MixerControls.CC.Read
    If \InteractiveMode Then
      Interpret MixerControls.CC.CLDisplayHandler CC", 1"

    /* Select channel mode */
    If RightSetup = "" Then
    Do
      /* Both channels */
      Channel = "B"
      Do J = 1 To Words(LeftSetup)
        Char = Word(LeftSetup, J)
        Interpret MixerControls.CC.CLKeyHandler
      End J
    End
    Else
    Do
      /* Separate channels */
      Channel = "L"
      Do J = 1 To Words(LeftSetup)
        Char = Word(LeftSetup, J)
        Interpret MixerControls.CC.CLKeyHandler
      End J
      Channel = "R"
      Do J = 1 To Words(RightSetup)
        Char = Word(RightSetup, J)
        Interpret MixerControls.CC.CLKeyHandler
      End J
    End

    /* Update mixer settings */
    Interpret MixerControls.CC.Write
    If \InteractiveMode Then
      Interpret MixerControls.CC.CLDisplayHandler CC", 0"
  End I
Return

/* Initialize CL data */
CLInit:
  /* Mixer controls CL data */
  MixerControls.1.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.1.CLDisplayHandler = "Call CLDrawNormalControl"
  MixerControls.2.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.2.CLDisplayHandler = "Call CLDrawNormalControl"
  MixerControls.3.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.3.CLDisplayHandler = "Call CLDrawNormalControl"
  MixerControls.4.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.4.CLDisplayHandler = "Call CLDrawNormalControl"
  MixerControls.5.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.5.CLDisplayHandler = "Call CLDrawNormalControl"
  MixerControls.6.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.6.CLDisplayHandler = "Call CLDrawNormalControl"
  MixerControls.7.CLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLRecordControl()"
  MixerControls.7.CLDisplayHandler = "Call CLDrawRecordControl"
Return

/* Propagate operation to required channels */
CLChannelOp:
  Parse Arg LValue1"%"LValue2, RValue
  Select
    When Channel = "L" Then
      Interpret LValue1 || "MixerState.Left" || LValue2 "=" RValue
    When Channel = "R" Then
      Interpret LValue1 || "MixerState.Right" || LValue2 "=" RValue
    Otherwise
      Interpret LValue1 || "MixerState.Left" || LValue2 "=" RValue
      Interpret LValue1 || "MixerState.Right" || LValue2 "=" RValue
  End
Return

/* CL params handlers */

/* Volume control CL handler */
CLVolumeControl:
  /* Check if there any numrcials */
  If Verify(Char, "0123456789", "M") \= 0 Then
    Select

      /* Precentage */
      When Right(Char, 1) = "%" Then
      Do
        /* Strip trailing "%" */
        Char = Strip(Char, "T", "%",)
        /* Skip if value isn't whole number */
        If \DataType(Char, "W") Then
          Return 0
        /* Bounds check */
        If Char < 0 Then
          Char = 0
        If Char > 100 Then
          Char = 100
        /* Update setting */
        Call CLChannelOp "%", "Format(MixerControls.CC.LevelMin + Char / 100 * (MixerControls.CC.LevelMax - MixerControls.CC.LevelMin),, 0)"
      End

      /* dB */
      When Right(Char, 2) = "DB" Then
      Do
        /* Strip trailing "dB" */
        Char = Left(Char, Length(Char) - 2)
        /* Skip if value isn't a number */
        If \DataType(Char, "N") Then
          Return 0
        /* Bounds check */
        If Char < MixerControls.CC.LevelValueMin Then
          Char = MixerControls.CC.LevelValueMin
        If Char > MixerControls.CC.LevelValueMax Then
          Char = MixerControls.CC.LevelValueMax
        /* Update setting */
        Call CLChannelOp "%", "Format(MixerControls.CC.LevelMin - (Char - MixerControls.CC.LevelValueMin) / MixerControls.CC.LevelValueStep,, 0)"
      End

      /* Absolute */
      Otherwise
        /* Skip if value isn't a whole number */
        If \DataType(Char, "W") Then
          Return 0
        /* Bounds check */
        If Char > MixerControls.CC.LevelMin Then
          Char = MixerControls.CC.LevelMin
        If Char < MixerControls.CC.LevelMax Then
          Char = MixerControls.CC.LevelMax
        /* Update setting */
        Call CLChannelOp "%", "Char"
    End
  Else
    /* No numericals - not a level */
    Return 1
Return 0

/* Mute control CL handler */
CLMuteControl: 
  Select
    /* Mute */
    When Char = "M" Then
      Call CLChannelOp "%_Mute", "1"
    /* Unmute */
    When Char = "U" Then
      Call CLChannelOp "%_Mute", "0"

    /* No params catched */
    Otherwise
      Return 1
  End
Return 0

/* Record control CL handler */
CLRecordControl: 
  Select
    /* Source = Line In */
    When Char = "L" Then
      Call CLChannelOp "%_Source", "0"
    /* Source = CD */
    When Char = "C" Then
      Call CLChannelOp "%_Source", "1"
    /* Source = Mic */
    When Char = "M" Then
      Call CLChannelOp "%_Source", "2"
    /* Source = Mix */
    When Char = "X" Then
      Call CLChannelOp "%_Source", "3"
    /* Mic +20dB gain on */
    When Char = "G" Then
      Call CLChannelOp "%_Mic20dBGain", "1"
    /* Mic +20dB gain off */
    When Char = "N" Then
      Call CLChannelOp "%_Mic20dBGain", "0"
    /* Mic +20dB gain toggle */

    /* No params catched */
    Otherwise
      Return 1
  End
Return 0

/* Draw mixer face */
CLDrawMixerFace: Procedure
  Call LineOut "CON", "    Control           Left            Right          Source         Mic Gain"
Return

/* Draw a standard mixer control */
CLDrawNormalControl: Procedure Expose MixerControls. MixerState.
  Arg Index, State

  If State Then
    Call CharOut "CON", " [1;37m" || Center(MixerControls.Index.Name, 13) || "[0;37m"
  Else
    Call CharOut "CON", " [1;37m" || Center("->", 13) || "[0;37m"

  Pos = MixerControls.Index.LevelMin - MixerState.Left
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", "   [1;" || 32 - MixerState.Left_Mute || "m" ||,
    Right(Level, 5) || "dB (" || Right(Format(100 * Pos / (MixerControls.Index.LevelMin - MixerControls.Index.LevelMax),,0), 3) || "%)[0;37m"

  Pos = MixerControls.Index.LevelMin - MixerState.Right
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call LineOut "CON", " / [1;" || 32 - MixerState.Right_Mute || "m" ||,
    Right(Level, 5) || "dB (" || Right(Format(100 * Pos / (MixerControls.Index.LevelMin - MixerControls.Index.LevelMax),,0), 3) || "%)[0;37m"
Return

/* Draw a standard mixer control */
CLDrawRecordControl: Procedure Expose MixerControls. MixerState.
  Arg Index, State

  If State Then
    Call CharOut "CON", " [1m" || Center(MixerControls.Index.Name, 13) || "[0m"
  Else
    Call CharOut "CON", " [1m" || Center("->", 13) || "[0m"
  Pos = MixerControls.Index.LevelMin - MixerState.Left
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", "   [1;32m" ||,
                      Right(Level, 5) || "dB (" || Right(Format(100 * Pos / (MixerControls.Index.LevelMin - MixerControls.Index.LevelMax),,0), 3) || "%)[0m"
  Pos = MixerControls.Index.LevelMin - MixerState.Right
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", " / [1;32m" ||,
                      Right(Level, 5) || "dB (" || Right(Format(100 * Pos / (MixerControls.Index.LevelMin - MixerControls.Index.LevelMax),,0), 3) || "%)[0m"
  Call LineOut "CON", "   [1m" ||, 
                      Center(Value("MixerControls.Index.SourcesShort." || MixerState.Left_Source), 4) ||,
                      "[0m / [1m" ||,
                      Center(Value("MixerControls.Index.SourcesShort." || MixerState.Right_Source), 4) ||,
                      "[0m   [1m" ||,
                      Center(Value("MixerControls."Index".Mic20dBGainStatesShort." || MixerState.Left_Mic20dBGain), 5) ||,
                      "[0m / [1m" ||,
                      Center(Value("MixerControls."Index".Mic20dBGainStatesShort." || MixerState.Right_Mic20dBGain), 5) || "[0m"
Return


/* User interface */

/* Interactive mode procedure */
UIProc:
  Arg ChangeWindowSize

  /* Initialize UI data */
  Call UIInit

  /* Draw mixer face */
  Call UIDrawMixerFace ChangeWindowSize

  /* Draw initial mixer state */
  Do CC = 1 To MixerControls.Num
    Interpret MixerControls.CC.Read
    Interpret MixerControls.CC.UIDisplayHandler CC
  End CC

  /* Draw selector */
  CC = 1
  NewCC = 1
  Interpret MixerControls.CC.UISelectorHandler CC", 1"

  /* Main loop */
  Do Forever

    /* Get current control state */
    Interpret MixerControls.CC.Read

    /* Redisplay control */
    Interpret MixerControls.CC.UIDisplayHandler CC

    /* Blocking input*/
    Char = SysGetKey("NOECHO")
    If Char = "E0"x | Char = "00"x Then
      Char = Char || (SysGetKey("NOECHO"))

    /* Handle control's internal keys */
    Interpret MixerControls.CC.UIKeyHandler

    /* Handle global keys */
    If NoMixerUpdate Then Do
      Select
        /* Exit */
        When Char = "1B"x Then
          Leave

        /* Selector up */
        When Char = "E048"x Then
          NewCC = Word(MixerControls.CC.UIJumpTable, 1)
        /* Selector down */
        When Char = "E050"x Then
        NewCC = Word(MixerControls.CC.UIJumpTable, 2)
        /* Selector left */
        When Char = "E04B"x Then
          NewCC = Word(MixerControls.CC.UIJumpTable, 3)
        /* Selector right */
        When Char = "E04D"x Then
          NewCC = Word(MixerControls.CC.UIJumpTable, 4)

        /* Refresh all controls */
        When Char = " " Then
          Do I = 1 To MixerControls.Num
            Interpret MixerControls.I.Read
            Interpret MixerControls.I.UIDisplayHandler I
          End I

        /* Any other keys */
        Otherwise
          NOP
      End
      /* Move selctor */
      If NewCC \= CC Then Do
        Interpret MixerControls.CC.UISelectorHandler CC", 0"
        CC = NewCC
        Interpret MixerControls.CC.UISelectorHandler CC", 1"
      End
    End
    Else
     Interpret MixerControls.CC.Write
  End

  /* Close mixer face */
  Call UICloseMixerFace
Return

/* Initialize UI data */
UIInit:
  /* Mixer controls UI data */
  MixerControls.1.UIPos = ""
  MixerControls.1.UIJumpTable = "6 3 2 2"
  MixerControls.1.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIMuteControl()"
  MixerControls.1.UICLKeyHandler = "NoMixerUpdate = CLVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = CLMuteControl()"
  MixerControls.1.UIDisplayHandler = "Call UIDrawNormal32LevelControl"
  MixerControls.1.UISelectorHandler = "Call UIDrawNormalSelector"
  MixerControls.2.UIPos = "[32C"
  MixerControls.2.UIJumpTable = "7 4 1 1"
  MixerControls.2.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIMuteControl()"
  MixerControls.2.UIDisplayHandler = "Call UIDrawNormal32LevelControl"
  MixerControls.2.UISelectorHandler = "Call UIDrawNormalSelector"
  MixerControls.3.UIPos = "[3B"
  MixerControls.3.UIJumpTable = "1 5 4 4"
  MixerControls.3.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIMuteControl()"
  MixerControls.3.UIDisplayHandler = "Call UIDrawNormal16LevelControl"
  MixerControls.3.UISelectorHandler = "Call UIDrawNormalSelector"
  MixerControls.4.UIPos = "[3B[32C"
  MixerControls.4.UIJumpTable = "2 7 3 3"
  MixerControls.4.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIMuteControl()"
  MixerControls.4.UIDisplayHandler = "Call UIDrawNormal16LevelControl"
  MixerControls.4.UISelectorHandler = "Call UIDrawNormalSelector"
  MixerControls.5.UIPos = "[6B"
  MixerControls.5.UIJumpTable = "3 6 7 7"
  MixerControls.5.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIMuteControl()"
  MixerControls.5.UIDisplayHandler = "Call UIDrawNormal16LevelControl"
  MixerControls.5.UISelectorHandler = "Call UIDrawNormalSelector"
  MixerControls.6.UIPos = "[9B"
  MixerControls.6.UIJumpTable = "5 1 7 7"
  MixerControls.6.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIMuteControl()"
  MixerControls.6.UIDisplayHandler = "Call UIDrawNormal16LevelControl"
  MixerControls.6.UISelectorHandler = "Call UIDrawNormalSelector"
  MixerControls.7.UIPos = "[6B[32C"
  MixerControls.7.UIJumpTable = "4 2 5 6"
  MixerControls.7.UIKeyHandler = "NoMixerUpdate = UIVolumeControl(); If NoMixerUpdate Then NoMixerUpdate = UIRecordControl()"
  MixerControls.7.UIDisplayHandler = "Call UIDrawRecordControl"
  MixerControls.7.UISelectorHandler = "Call UIDrawNormalSelector"
Return

/* Key handlers */

/* Volume control key handler */
UIVolumeControl: 
  /* Choose key */
  Select
    /* Left channel volume up */
    When (Char = "7" | Char = "0047"x) & MixerState.Left \= MixerControls.CC.LevelMax Then
      MixerState.Left = MixerState.Left - 1
    /* Left channel volume down */
    When (Char = "1" | Char = "004F"x) & MixerState.Left \= MixerControls.CC.LevelMin Then
      MixerState.Left = MixerState.Left + 1
    /* Left channel volume max */
    When Char = "0077"x Then
      MixerState.Left = MixerControls.CC.LevelMax
    /* Left channel volume min */
    When Char = "0075"x Then
      MixerState.Left = MixerControls.CC.LevelMin
    /* Right channel volume up */
    When (Char = "9" | Char = "0049"x) & MixerState.Right \= MixerControls.CC.LevelMax Then
      MixerState.Right = MixerState.Right - 1
    /* Right channel volume down */
    When (Char = "3" | Char = "0051"x) & MixerState.Right \= MixerControls.CC.LevelMin Then
      MixerState.Right = MixerState.Right + 1
    /* Right channel volume max */
    When Char = "0084"x Then
      MixerState.Right = MixerControls.CC.LevelMax
    /* Right channel volume min */
    When Char = "0076"x Then
      MixerState.Right = MixerControls.CC.LevelMin
    /* Both channels volume up */
    When Char = "8" | Char = "0048"x Then Do
      If MixerState.Left \= MixerControls.CC.LevelMax Then
        MixerState.Left = MixerState.Left - 1
      If MixerState.Right \= MixerControls.CC.LevelMax Then
        MixerState.Right = MixerState.Right - 1
    End
    /* Both channels volume down */
    When Char = "2" | Char = "0050"x Then Do
      If MixerState.Left \= MixerControls.CC.LevelMin Then
        MixerState.Left = MixerState.Left + 1
      If MixerState.Right \= MixerControls.CC.LevelMin Then
        MixerState.Right = MixerState.Right + 1
    End
    /* Both channels volume max */
    When Char = "008D"x Then Do
      MixerState.Left = MixerControls.CC.LevelMax
      MixerState.Right = MixerControls.CC.LevelMax
    End
    /* Both channels volume min */
    When Char = "0091"x Then Do
      MixerState.Left = MixerControls.CC.LevelMin
      MixerState.Right = MixerControls.CC.LevelMin
    End
    /* Right channel volume = left channel volume */
    When Char = "0" | Char = "0052"x Then Do
      MixerState.Right = MixerState.Left
      Return 1
    End
    /* Left channel volume = right channel volume */
    When Char = "." | Char = "0053"x Then Do
      MixerState.Left = MixerState.Right
      Return 1
    End
    /* No keys catched */
    Otherwise
      Return 1
  End
Return 0

/* Mute control key handler */
UIMuteControl: 
  /* Choose key */
  Select
    /* Left channel mute toggle */
    When Char = "4" | Char = "004B"x Then
      MixerState.Left_Mute = \ MixerState.Left_Mute
    /* Right channel mute toggle */
    When Char = "6" | Char = "004D"x Then
      MixerState.Right_Mute = \ MixerState.Right_Mute
    /* Both channels mute toggle */
    When Char = "5" | Char = "004C"x Then Do
      MixerState.Left_Mute = \ MixerState.Left_Mute
      MixerState.Right_Mute = \ MixerState.Right_Mute
    End
    /* Right channel mute = left channel mute */
    When Char = "0" | Char = "0052"x Then
      MixerState.Right_Mute = MixerState.Left_Mute
    /* Left channel mute = right channel mute */
    When Char = "." | Char = "0053"x Then
      MixerState.Left_Mute = MixerState.Right_Mute
    /* No keys catched */
    Otherwise
      Return 1
  End
Return 0

/* Record control key handler */
UIRecordControl: 
  /* Choose key */
  Select
    /* Left channel source increment */
    When Char = "4" | Char = "004B"x Then
      If MixerState.Left_Source = 3 Then
        MixerState.Left_Source = 0
      Else
        MixerState.Left_Source = MixerState.Left_Source + 1
    /* Right channel source increment */
    When Char = "6" | Char = "004D"x Then
      If MixerState.Right_Source = 3 Then
        MixerState.Right_Source = 0
      Else
        MixerState.Right_Source = MixerState.Right_Source + 1
    /* Both channels source increment */
    When Char = "5" | Char = "004C"x Then Do
      If MixerState.Left_Source = 3 Then
        MixerState.Left_Source = 0
      Else
        MixerState.Left_Source = MixerState.Left_Source + 1
      If MixerState.Right_Source = 3 Then
        MixerState.Right_Source = 0
      Else
        MixerState.Right_Source = MixerState.Right_Source + 1
    End
    /* Left channel +20dB gain toggle */
    When Char = "0073"x Then
      MixerState.Left_Mic20dBGain = \ MixerState.Left_Mic20dBGain
    /* Right channel +20dB gain toggle */
    When Char = "0074"x Then
      MixerState.Right_Mic20dBGain = \ MixerState.Right_Mic20dBGain
    /* Both channels +20dB gain toggle */
    When Char = "008F"x Then Do
      MixerState.Left_Mic20dBGain = \ MixerState.Left_Mic20dBGain
      MixerState.Right_Mic20dBGain = \ MixerState.Right_Mic20dBGain
    End
    /* Right channel parameters = left channel parameters */
    When Char = "0" | Char = "0052"x Then Do
      MixerState.Right_Source = MixerState.Left_Source
      MixerState.Right_Mic20dBGain = MixerState.Left_Mic20dBGain
    End
    /* Left channel parameters = right channel parameters */
    When Char = "." | Char = "0053"x Then Do
      MixerState.Left_Source = MixerState.Right_Source
      MixerState.Left_Mic20dBGain = MixerState.Right_Mic20dBGain
    End
    /* No keys catched */
    Otherwise
      Return 1
  End
Return 0

/* Display handlers */

/* Draw mixer face */
UIDrawMixerFace: Procedure
  Arg ChangeWindowSize

  If ChangeWindowSize Then
    "@MODE 67 14 2>NUL >NUL"
  Call SysCurState "OFF"
  Call CharOut "CON", "[=7l" ||,
    "[47;30m  OPTi 82C930 Mixer Ŀ [0m" || "0D0A"x ||,
    "[47;30m  Master Volume  Digital Audio Ĵ [0m" || "0D0A"x ||,
    "[47;30m  L: [                ]      dB  L: [                ]      dB  [0m" || "0D0A"x ||,
    "[47;30m  R: [                ]      dB  R: [                ]      dB  [0m" || "0D0A"x ||,
    "[47;30m  MIDI  Compact Disc Ĵ [0m" || "0D0A"x ||,
    "[47;30m  L: [                ]      dB  L: [                ]      dB  [0m" || "0D0A"x ||,
    "[47;30m  R: [                ]      dB  R: [                ]      dB  [0m" || "0D0A"x ||,
    "[47;30m  Line In  Record Control Ĵ [0m" || "0D0A"x ||,
    "[47;30m  L: [                ]      dB  L: [                ]      dB  [0m" || "0D0A"x ||,
    "[47;30m  R: [                ]      dB  R: [                ]      dB  [0m" || "0D0A"x ||,
    "[47;30m  Microphone Ĵ  Source  Mic Gain   [0m" || "0D0A"x ||,
    "[47;30m  L: [                ]      dB  L:               L:           [0m" || "0D0A"x ||,
    "[47;30m  R: [                ]      dB  R:               R:           [0m" || "0D0A"x ||,
    "[47;30m  [67D[11A[7C[47;30m"
Return

/* Close mixer face */
UICloseMixerFace: Procedure
  Call LineOut "CON", "[0m[10B"
  Call SysCurState "ON"
Return

/* Draw a single standard 32-level mixer control */
UIDrawNormal32LevelControl: Procedure Expose MixerControls. MixerState.
  Arg Index

  Call CharOut "CON", "[s" || MixerControls.Index.UIPos
  Pos = MixerControls.Index.LevelMin - MixerState.Left
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  If Pos // 2 Then
    Call CharOut "CON", Overlay("[1;" || 32 - MixerState.Left_Mute || "m[30m", "                            ", Pos % 2 + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[B[23D"
  Else
    Call CharOut "CON", Overlay("[1;" || 32 - MixerState.Left_Mute || "m[30m", "                            ", Pos % 2 + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[B[23D"
  Pos = MixerControls.Index.LevelMin - MixerState.Right
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  If Pos // 2 Then
    Call CharOut "CON", Overlay("[1;" || 32 - MixerState.Right_Mute || "m[30m", "                            ", Pos % 2 + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[u"
  Else
    Call CharOut "CON", Overlay("[1;" || 32 - MixerState.Right_Mute || "m[30m", "                            ", Pos % 2 + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[u"
Return

/* Draw a single standard 16-level mixer control */
UIDrawNormal16LevelControl: Procedure Expose MixerControls. MixerState.
  Arg Index

  Call CharOut "CON", "[s" || MixerControls.Index.UIPos
  Pos = MixerControls.Index.LevelMin - MixerState.Left
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", Overlay("[1;" || 32 - MixerState.Left_Mute || "m[30m", "                            ", Pos + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[B[23D"
  Pos = MixerControls.Index.LevelMin - MixerState.Right
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", Overlay("[1;" || 32 - MixerState.Right_Mute || "m[30m", "                            ", Pos + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[u"
Return

/* Draw a standard mixer control selector */
UIDrawNormalSelector: Procedure Expose MixerControls.
  Arg Index, State

  If State Then
    Call CharOut "CON", "[s" || MixerControls.Index.UIPos || "[A[5D" || Left("" MixerControls.Index.Name "", 31) || "[u"
  Else
    Call CharOut "CON", "[s" || MixerControls.Index.UIPos || "[A[5D" || Left("" MixerControls.Index.Name "", 31) || "[u"
Return 

/* Draw a record mixer control */
UIDrawRecordControl: Procedure Expose MixerControls. MixerState.
  Arg Index

  Call CharOut "CON", "[s" || MixerControls.Index.UIPos
  Pos = MixerControls.Index.LevelMin - MixerState.Left
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", Overlay("[1;32m[30m", "                            ", Pos + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[B[23D"
  Pos = MixerControls.Index.LevelMin - MixerState.Right
  Level = Pos * MixerControls.Index.LevelValueStep + MixerControls.Index.LevelValueMin
  If Level > 0 Then Level = "+" || Level
  Call CharOut "CON", Overlay("[1;32m[30m", "                            ", Pos + 1) || "[0;47;30m[2C" || Right(Level, 5) || "[2B[24D"
  Call CharOut "CON", Left(Value("MixerControls.Index.Sources." || MixerState.Left_Source), 12) || "[6C" ||Left(Value("MixerControls."Index".Mic20dBGainStates." || MixerState.Left_Mic20dBGain), 8) || "[B[26D" ||,
    Left(Value("MixerControls.Index.Sources." || MixerState.Right_Source), 12) || "[6C" ||Left(Value("MixerControls."Index".Mic20dBGainStates." || MixerState.Right_Mic20dBGain), 8) || "[u"
Return


/* High-level mixer functions */

/* Check for hardware & initialize data structures */
MixerInit: Procedure Expose Mixer_MWSSPort MixerControls.

  /* Check for & initialize hardware */
  If \OPTi930Init() Then
    Return 0

  /* Mixer controls data */
  MixerControls.Num = 7
  MixerControls.1.Name = "Master Volume"
  MixerControls.1.Abbrev = "MAIN"
  MixerControls.1.LevelMin = 31
  MixerControls.1.LevelMax = 0
  MixerControls.1.LevelValueMin = -46.5
  MixerControls.1.LevelValueMax = 0
  MixerControls.1.LevelValueStep = 1.5
  MixerControls.1.Read = "Temp = MixerRead(22); MixerState.Left = Temp // 32; MixerState.Left_Mute = Temp % 128; " ||,
    "Temp = MixerRead(23); MixerState.Right = Temp // 32; MixerState.Right_Mute = Temp % 128"
  MixerControls.1.Write = "Call MixerWrite 22, MixerState.Left_Mute * 128 + MixerState.Left; " ||,
    "Call MixerWrite 23, MixerState.Right_Mute * 128 + MixerState.Right"
  MixerControls.2.Name = "Digital Audio"
  MixerControls.2.Abbrev = "WAVE"
  MixerControls.2.LevelMin = 31
  MixerControls.2.LevelMax = 0
  MixerControls.2.LevelValueMin = -93
  MixerControls.2.LevelValueMax = 0
  MixerControls.2.LevelValueStep = 3
  MixerControls.2.Read = "Temp = MixerRead(6); MixerState.Left = (Temp // 64) % 2; MixerState.Left_Mute = Temp % 128; " ||,
    "Temp = MixerRead(7); MixerState.Right = (Temp // 64) % 2; MixerState.Right_Mute = Temp % 128"
  MixerControls.2.Write = "Call MixerWrite 6, MixerState.Left_Mute * 128 + MixerState.Left * 2; " ||,
    "Call MixerWrite 7, MixerState.Right_Mute * 128 + MixerState.Right * 2"
  MixerControls.3.Name = "MIDI"
  MixerControls.3.Abbrev = "MIDI"
  MixerControls.3.LevelMin = 15
  MixerControls.3.LevelMax = 0
  MixerControls.3.LevelValueMin = -33
  MixerControls.3.LevelValueMax = 12
  MixerControls.3.LevelValueStep = 3
  MixerControls.3.Read = "Temp = MixerRead(4); MixerState.Left = (Temp // 32) % 2; MixerState.Left_Mute = Temp % 128; " ||,
    "Temp = MixerRead(5); MixerState.Right = (Temp // 32) % 2; MixerState.Right_Mute = Temp % 128"
  MixerControls.3.Write = "Call MixerWrite 4, MixerState.Left_Mute * 128 + MixerState.Left * 2; " ||,
    "Call MixerWrite 5, MixerState.Right_Mute * 128 + MixerState.Right * 2"
  MixerControls.4.Name = "Compact Disc"
  MixerControls.4.Abbrev = "CD"
  MixerControls.4.LevelMin = 15
  MixerControls.4.LevelMax = 0
  MixerControls.4.LevelValueMin = -33
  MixerControls.4.LevelValueMax = 12
  MixerControls.4.LevelValueStep = 3
  MixerControls.4.Read = "Temp = MixerRead(2); MixerState.Left = (Temp // 32) % 2; MixerState.Left_Mute = Temp % 128; " ||,
    "Temp = MixerRead(3); MixerState.Right = (Temp // 32) % 2; MixerState.Right_Mute = Temp % 128"
  MixerControls.4.Write = "Call MixerWrite 2, MixerState.Left_Mute * 128 + MixerState.Left * 2; " ||,
    "Call MixerWrite 3, MixerState.Right_Mute * 128 + MixerState.Right * 2"
  MixerControls.5.Name = "Line In"
  MixerControls.5.Abbrev = "LINE"
  MixerControls.5.LevelMin = 15
  MixerControls.5.LevelMax = 0
  MixerControls.5.LevelValueMin = -33
  MixerControls.5.LevelValueMax = 12
  MixerControls.5.LevelValueStep = 3
  MixerControls.5.Read = "Temp = MixerRead(18); MixerState.Left = (Temp // 32) % 2; MixerState.Left_Mute = Temp % 128; " ||,
    "Temp = MixerRead(19); MixerState.Right = (Temp // 32) % 2; MixerState.Right_Mute = Temp % 128"
  MixerControls.5.Write = "Call MixerWrite 18, MixerState.Left_Mute * 128 + MixerState.Left * 2; " ||,
    "Call MixerWrite 19, MixerState.Right_Mute * 128 + MixerState.Right * 2"
  MixerControls.6.Name = "Microphone"
  MixerControls.6.Abbrev = "MIC"
  MixerControls.6.LevelMin = 15
  MixerControls.6.LevelMax = 0
  MixerControls.6.LevelValueMin = -33
  MixerControls.6.LevelValueMax = 12
  MixerControls.6.LevelValueStep = 3
  MixerControls.6.Read = "Temp = MixerRead(20); MixerState.Left = Temp // 16; MixerState.Left_Mute = Temp % 128; " ||,
    "Temp = MixerRead(21); MixerState.Right = Temp // 16; MixerState.Right_Mute = Temp % 128"
  MixerControls.6.Write = "Call MixerWrite 20, MixerState.Left_Mute * 128 + MixerState.Left; " ||,
    "Call MixerWrite 21, MixerState.Right_Mute * 128 + MixerState.Right"
  MixerControls.7.Name = "Recording"
  MixerControls.7.Sources.0 = "Line In"
  MixerControls.7.Sources.1 = "Compact Disc"
  MixerControls.7.Sources.2 = "Microphone"
  MixerControls.7.Sources.3 = "Mixer"
  MixerControls.7.SourcesShort.0 = "Line"
  MixerControls.7.SourcesShort.1 = "CD"
  MixerControls.7.SourcesShort.2 = "Mic"
  MixerControls.7.SourcesShort.3 = "Mix"
  MixerControls.7.Mic20dBGainStates.0 = "Disabled"
  MixerControls.7.Mic20dBGainStates.1 = "+20dB"
  MixerControls.7.Mic20dBGainStatesShort.0 = "Off"
  MixerControls.7.Mic20dBGainStatesShort.1 = "+20dB"
  MixerControls.7.Abbrev = "REC"
  MixerControls.7.LevelMin = 15
  MixerControls.7.LevelMax = 0
  MixerControls.7.LevelValueMin = 0
  MixerControls.7.LevelValueMax = 22.5
  MixerControls.7.LevelValueStep = 1.5
  MixerControls.7.Read = "Temp = MixerRead(0); MixerState.Left = Temp // 16; MixerState.Left_Mic20dBGain = (Temp // 64) % 32; MixerState.Left_Source = Temp % 64; " ||,
    "Temp = MixerRead(1); MixerState.Right = Temp // 16; MixerState.Right_Mic20dBGain = (Temp // 64) % 32; MixerState.Right_Source = Temp % 64"
  MixerControls.7.Write = "Call MixerWrite 0, MixerState.Left_Source * 64 + MixerState.Left_Mic20dBGain * 32 + MixerState.Left; " ||,
    "Call MixerWrite 1, MixerState.Right_Source * 64 + MixerState.Right_Mic20dBGain * 32 + MixerState.Right"
Return 1

/* Write to mixer */
MixerWrite: Procedure Expose Mixer_MWSSPort
  Arg Device, Value

  /* Enable CS4231's extended mode (mode2) */
  State = OPTi930Read(5)
  Call OPTi930Write 5,  (State % 64) * 64 + 32 + State // 32

  /* Write to MWSS mixer */
  Call RxOutPort Mixer_MWSSPort + 4, Device
  Call RxOutPort Mixer_MWSSPort + 5, Value

  /* Restore OPTi930 state */
  Call OPTi930Write 5, State
Return

/* Read from mixer */
MixerRead: Procedure Expose Mixer_MWSSPort
  Arg Device

  /* Enable CS4231's extended mode (mode2) */
  State = OPTi930Read(5)
  Call OPTi930Write 5, (State % 64) * 64 + 32 + State // 32

  /* Read from MWSS mixer */
  RC = RxInOutPort(Mixer_MWSSPort + 4, Device, Mixer_MWSSPort + 5)

  /* Restore OPTi930 state */
  Call OPTi930Write 5, State
Return RC


/* Low-level mixer functions */

/* Check for OPTi930 & get MWSS base port */
OPTi930Init: Procedure Expose Mixer_MWSSPort

  /* MWSS base ports */
  MWSSPorts.0 = 1328
  MWSSPorts.1 = 3712
  MWSSPorts.2 = 3904
  MWSSPorts.3 = 1540

  /* Check for OPTi930 */
  Temp = OPTi930Read(7)
  Call OPTi930Write 7, 203
  Pattern = OPTi930Read(7)
  Call OPTi930Write 7, Temp

  /* Determine MWSS port if OPTi930 present */
  If Pattern = 203 Then
  Do
    Mixer_MWSSPort = Value("MWSSPorts." || (OPTi930Read(1) // 64) % 16)
    Return 1
  End
Return 0

/* Write to OPTi930 */
OPTi930Write: Procedure
  Arg Index, Value

  /* Unlock OPTi930 */
  Call RxOutPort 3983, 128

  /* Write to OPTi930 register */
  Call RxOutPortW 3598, Index + Value * 256
Return

/* Read from OPTi930 */
OPTi930Read: Procedure
  Arg Index

  /* Unlock OPTi930 */
  Call RxOutPort 3983, 128

  /* Read from OPTi930 register */
Return RxInOutPort(3598, Index, 3599)


/* Signal handlers */

/* Disable Ctrl-Break */
CtrlBreak:
  Signal On Halt Name CtrlBreak
Return

/* End of Mixer930.CMD */
